# frozen_string_literal: true

TelegramWebProxy::MessageProcessor = Struct.new(:bot, :chat) do
  MAX_FRAGMENT_SIZE = 4096
  WEBSEARCH_URL = 'https://duckduckgo.com/lite'
  HELP_TEXT =
    <<~MSG
      Supported requests:
      /h - help
      /o URL (or just URL without /o) - open URL
      /s REQUEST - search with DuckDuckGo
      /e - echo
      /t - just a test
    MSG
  # /f is possible too

  TEST_TEXT = "A ! <a href='tg:msg?text=test&to=wprox_bot'> click here</a> <a href='https://techcrunch.com/2017/07/14/startup-battlefield-application-deadline-extended-for-disrupt-sf-2/'>TC instant view</a> <a href='http://reddit.com'>Reddit</a>." # rubocop:disable Metrics/LineLength

  attr_accessor :message, :page

  def process(message)
    self.message = message
    if message.instance_of? Telegram::Bot::Types::CallbackQuery
      process_callback
    else
      process_message
    end
  end

  def process_callback
    command, args = message.data.split ' '
    case command
    when '/f'
      send_fragment(args.to_i)
    end
  end

  def process_message # rubocop:disable Metrics/MethodLength, Metrics/CyclomaticComplexity
    command, args = message.text.split(' ', 2)
    case command
    when '/h'
      send_text(text: HELP_TEXT)
    when '/o'
      open_page(args)
    when '/s'
      web_search(args)
    when '/e'
      send_text(text: args)
    when '/f' # actually user shouldn't use this manually, see #process_callback
      send_fragment(args.to_i)
    when '/'
      puts 'Do nothing'
    when '/t'
      send_text(text: TEST_TEXT)
    else
      process_unknown_command(command)
    end
  end

  private

  def looks_like_url?(str)
    str =~ URI.regexp ||
      str =~ /\A[\w-]+\.[\w-]+(\.[\w-]+)*\z/
  end

  # FIXME: check if given command is looks like an url
  def process_unknown_command(command)
    unless looks_like_url?(command)
      send_text(text: 'Not a command or URL')
    else
      puts "Trying to open page: #{command}"
      send_object(page_fetcher(command).get)
    end
  end

  def open_page(args)
    send_object(page_fetcher(args).get)
  end

  def web_search(args)
    send_object(page_fetcher(WEBSEARCH_URL).post(q: args))
  end

  def send_text(text:, reply_markup: nil, parse_mode: 'HTML')
    bot.api.send_message(chat_id: chat.id,
                         text: text,
                         parse_mode: parse_mode,
                         reply_markup: reply_markup)
  end

  # Sends a fragment of the previously fetched web page to specified chat
  #
  # @param chat_id [Integer]
  # @param fragment [Integer] the number of the fragment to send
  #
  def send_fragment(fragment)
    keyboard_markup =
      TelegramWebProxy::KeyboardGenerator.new(page.size, fragment).generate
    begin
      send_text(text: page[fragment], reply_markup: keyboard_markup)
    rescue Telegram::Bot::Exceptions::ResponseError => e
      handle_response_error(e, fragment, keyboard_markup)
    end
  end

  def handle_response_error(e, fragment, keyboard_markup)
    if e.error_code.to_s == '400'
      send_text(
        text: "Error with fragment #{fragment + 1}. Previous one:\n" + page[fragment - 1],
        reply_markup: keyboard_markup
      )
    else
      send_text(text: "Error while sending fragment #{fragment + 1}")
    end
  end

  def split_into_parts(page)
    return [] if page.empty?
    size = [page.size, MAX_FRAGMENT_SIZE].min
    size = page[0, size].rindex('<') until valid_segment?(page[0, size])
    raise "INVALID FRAGMENT near '#{page[0..500]}...'" unless size.positive?
    [page[0, size], *split_into_parts(page[size..-1])]
  end

  # Checks if the segment of the page is valid and might be taken as a separate fragment
  def valid_segment?(page_segment)
    Nokogiri::XML("<e>#{page_segment}</e>").errors.empty?
  end

  def send_object(rest_response)
    case rest_response.headers[:content_type].split('/')[0]
    when 'text'
      split_and_send(TelegramWebProxy::PageSanitizer.new(rest_response).sanitize)
    when 'image'
      send_text(text: '^')
    else
      send_text(text: 'Only text or image content is supported')
    end
  end

  # Splits fetched HTML and sends the first fragment
  def split_and_send(html_doc)
    self.page = split_into_parts(html_doc)
    send_fragment(0)
  end

  def page_fetcher(url)
    TelegramWebProxy::PageFetcher.new(url)
  end
end
