# coding: utf-8
# frozen_string_literal: true

TelegramWebProxy::PageSanitizer = Struct.new(:response) do
  WHITESPACE_ELEMENTS = {'br'  => {before: "\n", after: ''},
                         'div' => {before: "\n", after: "\n"},
                         'p'   => {before: "\n", after: "\n"}}.freeze

  def sanitize
    fix_relative_urls(prepare_page(response.body), response.request.url)
  end

  private

  # Takes out most of HTML tags and unnecesary whitespaces
  def prepare_page(body)
    remove_whitespaces(transform_img(remove_nested_tags(sanitize_tags(body))))
  end

  def image_description(attrs)
    CGI.escapeHTML((attrs['alt'] || attrs['title'] || 'img').to_s[0..20])
  end

  def image_replacement(attrs)
    return '' if attrs['width'].to_s == '1' && attrs['height'].to_s == '1'
    "<a href=\"#{CGI.escapeHTML(attrs['src'])}\">〖#{image_description(attrs)}〗</a>"
  end

  # Replaces <img src=... alt=...> with <a href=img_src>img_alt</a>
  def transform_img(text)
    replacements = text.scan(/<\\?img [^>]*>/).map do |img_tag|
      attrs = Nokogiri::HTML.fragment(img_tag).children[0].attributes
      [img_tag, image_replacement(attrs)]
    end
    replacements.inject(text) { |str, tags| str.gsub(tags[0], tags[1]) }
  end

  def remove_nested_tags(text)
    tree = Nokogiri::HTML.fragment(text)
    [process_tree(tree)].flatten.map(&:to_s).join
  end

  def no_nested_tags?(tree, tag_name)
    tree.children.none? { |c| c.enum_for(:traverse).map(&:name).include?(tag_name) }
  end

  # Makes the tree flat.
  # It takes the top-level tag and remove all nested tags unless it has nested <a> tag
  # (otherwise it takes <a> tag and remove all wrapping tags)
  def process_tree(tree)
    return tree if tree.instance_of? Nokogiri::XML::Text
    if (tree.name == 'a' || no_nested_tags?(tree, 'a')) &&
       !tree.instance_of?(Nokogiri::HTML::DocumentFragment)
      text = tree.xpath('.//text()').to_s
      tree.children.map(&:remove)
      tree.add_child text
      tree
    else
      tree.children.map { |child| process_tree(child) }
    end
  end

  def sanitize_tags(body)
    Sanitize.fragment(
      body,
      elements: %w[b strong i em a code pre img],
      attributes: {'a' => ['href'], 'img' => %w[src alt title width height]},
      remove_contents: %i[script style img],
      whitespace_elements: WHITESPACE_ELEMENTS
    )
  end

  def remove_whitespaces(text)
    # FIXME: remove also linebreak from  "\n</tag>\n" and "\n<tag>\n"
    text.
      gsub(/\t+/, ' ').
      gsub(/ +/, ' ').
      gsub(/( \n)+/, "\n").
      gsub(/\n+/, "\n").
      strip
  end

  # Replaces relative urls to absolute.
  #
  # @param text [String] original text
  # @param url [String] url from where the page was fetched
  # @return [String]
  def fix_relative_urls(text, url)
    uri = URI.parse(url)
    text.gsub('ref="//', "ref=\"#{uri.scheme || 'http'}://").
      gsub('ref="/', "ref=\"#{uri.scheme || 'http'}://#{uri.host || uri.path}/")
  end
end
